/*
V.1.0 Beta
The code is the basic structure for the 3EE, tiny guests device. 
Further analysis and development should be considered.
The basis of the code is based mainly on the free software PicoRecPlayAudio - Raspberry Pi Pico Audio Record/Playback/WAV File
// Copyright (C) 2022 Jason Birch, along with many other codes on the Arduino forum, library examples, and help through ChatGPT.
All the knowledge about sound sampling and processing was acquired through online forums such as Arduino forum threads.

The code was designed to work on a Raspberry Pico Lipo, compiling the file through Arduino IDE. 
To do so is required to install from the board manager "Arduino Mbed OS Nano Boards" along with the SD and SPI libraries from the library manager. 

The code records audio for 9 seconds at 12000 Hz every set time and uploads the saved file to an SD card.
Further development and improvements:
-improve the WAV file generation
-increase the recording quality
-analyze the data collected and discard if relevant data is not identified instead of copying to SD
-deep sleep and low consumption when not recording

To better simplify the code and the data collection, a MEMS microphone could be used, instead of an Electric, along with CircuitPython, since those digital microphones 
don't require precise control over the ADC for sampling.
*/

#include <SD.h>
#include <SPI.h>

#define WAV_SAMPLE_RATE 12000                                               //Hz
#define AUDIO_CHANNELS 1                                                   //Mono
#define AUDIO_PERIOD 8                                                   //Recording duration in seconds
#define AUDIO_BUFF_SIZE (AUDIO_CHANNELS * WAV_SAMPLE_RATE * AUDIO_PERIOD)  //Buffer size to store sampling
#define WAV_PWM_COUNT (125000000 / WAV_SAMPLE_RATE)
#define SD_MOSI_PIN 19    // MOSI pin
#define SD_MISO_PIN 18    // MISO pin
#define SD_SCK_PIN 13     // SCK pin
const int CS_PIN = 17;    //CS pin
const int AUDIO_IN = A0;  //analog input
unsigned short AudioBuffer[AUDIO_BUFF_SIZE];
int number = 0;
char number1 = number;
String filenum = "01";
String filename = "Rec" + filenum + ".wav"; //first file name is data01.TXT

char header[] = {
    'R', 'I', 'F', 'F', // ChunkID
    0, 0, 0, 0,           // ChunkSize (to be filled later)
    'W', 'A', 'V', 'E', // Format
    'f', 'm', 't', ' ', // Subchunk1ID
    16, 0, 0, 0,         // Subchunk1Size (16 for PCM)
    1, 0,                // AudioFormat (PCM = 1)
    1, 0,                // NumChannels (Mono = 1, Stereo = 2)
    0, 0, 0, 0,          // SampleRate (to be filled later)
    0, 0, 0, 0,          // ByteRate (to be filled later)
    2, 0,                // BlockAlign (NumChannels * BitsPerSample / 8)
    16, 0,               // BitsPerSample (12-bit ADC samples stored as 16-bit)
    'd', 'a', 't', 'a', // Subchunk2ID
    0, 0, 0, 0           // Subchunk2Size (to be filled later)
};

File WAV_FILE;  //initialize file element,

//timer counter
unsigned long START_TIME;
unsigned long NOW_TIME;
const unsigned long RECORD_TIMER = 900000;  //a value in milliseconds

void setup() {
  // Open serial communications and wait for port to open:
  Serial.begin(115200);
  // wait for Serial Monitor to connect. Needed for native USB port boards only:
  while (!Serial);
  Serial.print("Initializing SD card...");

  if (!SD.begin(CS_PIN)) {
    Serial.println("initialization failed. Things to check:");
    Serial.println("1. is a card inserted?");
    Serial.println("2. is your wiring correct?");
    Serial.println("3. did you change the chipSelect pin to match your shield or module?");
    Serial.println("Note: press reset button on the board and reopen this Serial Monitor after fixing your issue!");
    while (true);
  }
  Serial.println("initialization done.");
  pinMode(AUDIO_IN, INPUT);  //initialize A0 input
  START_TIME = millis();     //start counter
  
//write file header
  writeWavHeader(WAV_FILE);
}

void loop() {
  //1
  NOW_TIME = millis();
  //2
  while (SD.exists(filename)) {
  int file_increment = filenum.toInt(); //change file increment to an integer
  ++file_increment; //add one to the file increment
  filenum = "0" + String(file_increment); //create a new increment with leading 0
  filename = "Rec" + filenum + ".wav";} //creates the new file name incremented by 1
  if (NOW_TIME - START_TIME >= RECORD_TIMER) {
    Serial.println("Recording start");
    AudioCapture(AudioBuffer);    
    // Open the WAV file for writing (create if it doesn't exist)
    WAV_FILE = SD.open(filename, FILE_WRITE);
    if (WAV_FILE) {
      Serial.print("Writing to test.wav...");
      // Write the sample to the WAV file
      unsigned long Headcount = 0;
      while (Headcount < sizeof(header)){
      WAV_FILE.print(header[Headcount++]);
      }
      unsigned long DumpCount = 0;
      const unsigned long Dump_Timer = 50; 
      while (DumpCount < AUDIO_BUFF_SIZE){
         unsigned long DumpStartMillis = millis(); //a value in milliseconds 
         while (DumpStartMillis - START_TIME <= Dump_Timer);
         WAV_FILE.print(AudioBuffer[DumpCount++]);
      }
      WAV_FILE.close();
      Serial.println("done.");
      number++;
    } else {
      // if the file didn't open, print an error:
      Serial.println("error opening file.wav");
    }
    Serial.println("Recording finished.");
    START_TIME = millis();
  }
 }

void AudioCapture(unsigned short AudioBuffer[]) {
  unsigned long SampleCount = 0;
  unsigned long long SamplePeriod;

  SampleCount = 0;
  AudioBuffer[SampleCount++] = (AUDIO_BUFF_SIZE & 0xFFFF);
  AudioBuffer[SampleCount++] = (AUDIO_BUFF_SIZE >> 16);

  SamplePeriod = time_us_64() + (1000000 / WAV_SAMPLE_RATE);
  while (SampleCount < AUDIO_BUFF_SIZE) {
    while (time_us_64() < SamplePeriod);
    SamplePeriod = time_us_64() + (1000000 / WAV_SAMPLE_RATE);
    AudioBuffer[SampleCount++] = WAV_PWM_COUNT * analogRead(AUDIO_IN) / 4096;
  }
}

void writeWavHeader(File &file) {
  // Fill in ChunkSize and Subchunk2Size
uint32_t chunkSize = sizeof(header) - 8 + AUDIO_BUFF_SIZE * sizeof(uint16_t); // 8 bytes for ChunkID and ChunkSize
uint32_t subchunk2Size = AUDIO_BUFF_SIZE * sizeof(uint16_t);                 // 2 bytes per sample (16-bit)

memcpy(&header[4], &chunkSize, sizeof(chunkSize));
memcpy(&header[40], &subchunk2Size, sizeof(subchunk2Size));

// Fill in SampleRate and ByteRate
uint32_t byteRate = WAV_SAMPLE_RATE * 1 * 16 / 8; // NumChannels * SampleRate * BitsPerSample / 8
uint32_t sampleRate = WAV_SAMPLE_RATE;

memcpy(&header[24], &sampleRate, sizeof(sampleRate));
memcpy(&header[28], &byteRate, sizeof(byteRate));
}

